# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: src/ray/protobuf/job_agent.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import agent_manager_pb2 as src_dot_ray_dot_protobuf_dot_agent__manager__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='src/ray/protobuf/job_agent.proto',
  package='ray.rpc',
  syntax='proto3',
  serialized_options=b'\370\001\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n src/ray/protobuf/job_agent.proto\x12\x07ray.rpc\x1a$src/ray/protobuf/agent_manager.proto\"B\n\x17InitializeJobEnvRequest\x12\'\n\x0fjob_description\x18\x01 \x01(\tR\x0ejobDescription\"\x8c\x01\n\x15InitializeJobEnvReply\x12/\n\x06status\x18\x01 \x01(\x0e\x32\x17.ray.rpc.AgentRpcStatusR\x06status\x12#\n\rerror_message\x18\x02 \x01(\tR\x0c\x65rrorMessage\x12\x1d\n\ndriver_pid\x18\x03 \x01(\x05R\tdriverPid2g\n\x0fJobAgentService\x12T\n\x10InitializeJobEnv\x12 .ray.rpc.InitializeJobEnvRequest\x1a\x1e.ray.rpc.InitializeJobEnvReplyB\x03\xf8\x01\x01\x62\x06proto3'
  ,
  dependencies=[src_dot_ray_dot_protobuf_dot_agent__manager__pb2.DESCRIPTOR,])




_INITIALIZEJOBENVREQUEST = _descriptor.Descriptor(
  name='InitializeJobEnvRequest',
  full_name='ray.rpc.InitializeJobEnvRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='job_description', full_name='ray.rpc.InitializeJobEnvRequest.job_description', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='jobDescription', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=83,
  serialized_end=149,
)


_INITIALIZEJOBENVREPLY = _descriptor.Descriptor(
  name='InitializeJobEnvReply',
  full_name='ray.rpc.InitializeJobEnvReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='status', full_name='ray.rpc.InitializeJobEnvReply.status', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='status', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='error_message', full_name='ray.rpc.InitializeJobEnvReply.error_message', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='errorMessage', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='driver_pid', full_name='ray.rpc.InitializeJobEnvReply.driver_pid', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='driverPid', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=152,
  serialized_end=292,
)

_INITIALIZEJOBENVREPLY.fields_by_name['status'].enum_type = src_dot_ray_dot_protobuf_dot_agent__manager__pb2._AGENTRPCSTATUS
DESCRIPTOR.message_types_by_name['InitializeJobEnvRequest'] = _INITIALIZEJOBENVREQUEST
DESCRIPTOR.message_types_by_name['InitializeJobEnvReply'] = _INITIALIZEJOBENVREPLY
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

InitializeJobEnvRequest = _reflection.GeneratedProtocolMessageType('InitializeJobEnvRequest', (_message.Message,), {
  'DESCRIPTOR' : _INITIALIZEJOBENVREQUEST,
  '__module__' : 'src.ray.protobuf.job_agent_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.InitializeJobEnvRequest)
  })
_sym_db.RegisterMessage(InitializeJobEnvRequest)

InitializeJobEnvReply = _reflection.GeneratedProtocolMessageType('InitializeJobEnvReply', (_message.Message,), {
  'DESCRIPTOR' : _INITIALIZEJOBENVREPLY,
  '__module__' : 'src.ray.protobuf.job_agent_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.InitializeJobEnvReply)
  })
_sym_db.RegisterMessage(InitializeJobEnvReply)


DESCRIPTOR._options = None

_JOBAGENTSERVICE = _descriptor.ServiceDescriptor(
  name='JobAgentService',
  full_name='ray.rpc.JobAgentService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=294,
  serialized_end=397,
  methods=[
  _descriptor.MethodDescriptor(
    name='InitializeJobEnv',
    full_name='ray.rpc.JobAgentService.InitializeJobEnv',
    index=0,
    containing_service=None,
    input_type=_INITIALIZEJOBENVREQUEST,
    output_type=_INITIALIZEJOBENVREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_JOBAGENTSERVICE)

DESCRIPTOR.services_by_name['JobAgentService'] = _JOBAGENTSERVICE

# @@protoc_insertion_point(module_scope)
