# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: src/ray/protobuf/event.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='src/ray/protobuf/event.proto',
  package='ray.rpc',
  syntax='proto3',
  serialized_options=b'\370\001\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1csrc/ray/protobuf/event.proto\x12\x07ray.rpc\"\xaa\x04\n\x05\x45vent\x12\x19\n\x08\x65vent_id\x18\x01 \x01(\tR\x07\x65ventId\x12:\n\x0bsource_type\x18\x02 \x01(\x0e\x32\x19.ray.rpc.Event.SourceTypeR\nsourceType\x12\'\n\x0fsource_hostname\x18\x03 \x01(\tR\x0esourceHostname\x12\x1d\n\nsource_pid\x18\x04 \x01(\x05R\tsourcePid\x12\x33\n\x08severity\x18\x05 \x01(\x0e\x32\x17.ray.rpc.Event.SeverityR\x08severity\x12\x14\n\x05label\x18\x06 \x01(\tR\x05label\x12\x18\n\x07message\x18\x07 \x01(\tR\x07message\x12\x1c\n\ttimestamp\x18\x08 \x01(\x03R\ttimestamp\x12\x45\n\rcustom_fields\x18\t \x03(\x0b\x32 .ray.rpc.Event.CustomFieldsEntryR\x0c\x63ustomFields\x1a?\n\x11\x43ustomFieldsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\">\n\nSourceType\x12\n\n\x06\x43OMMON\x10\x00\x12\x0f\n\x0b\x43ORE_WORKER\x10\x01\x12\x07\n\x03GCS\x10\x02\x12\n\n\x06RAYLET\x10\x03\"7\n\x08Severity\x12\x08\n\x04INFO\x10\x00\x12\x0b\n\x07WARNING\x10\x01\x12\t\n\x05\x45RROR\x10\x02\x12\t\n\x05\x46\x41TAL\x10\x03\"6\n\x11ReportEventsReply\x12!\n\x0csend_success\x18\x01 \x01(\x08R\x0bsendSuccess\":\n\x13ReportEventsRequest\x12#\n\revent_strings\x18\x01 \x03(\tR\x0c\x65ventStrings2^\n\x12ReportEventService\x12H\n\x0cReportEvents\x12\x1c.ray.rpc.ReportEventsRequest\x1a\x1a.ray.rpc.ReportEventsReplyB\x03\xf8\x01\x01\x62\x06proto3'
)



_EVENT_SOURCETYPE = _descriptor.EnumDescriptor(
  name='SourceType',
  full_name='ray.rpc.Event.SourceType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='COMMON', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CORE_WORKER', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='GCS', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='RAYLET', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=477,
  serialized_end=539,
)
_sym_db.RegisterEnumDescriptor(_EVENT_SOURCETYPE)

_EVENT_SEVERITY = _descriptor.EnumDescriptor(
  name='Severity',
  full_name='ray.rpc.Event.Severity',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='INFO', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='WARNING', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ERROR', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FATAL', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=541,
  serialized_end=596,
)
_sym_db.RegisterEnumDescriptor(_EVENT_SEVERITY)


_EVENT_CUSTOMFIELDSENTRY = _descriptor.Descriptor(
  name='CustomFieldsEntry',
  full_name='ray.rpc.Event.CustomFieldsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='ray.rpc.Event.CustomFieldsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='key', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='ray.rpc.Event.CustomFieldsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='value', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=412,
  serialized_end=475,
)

_EVENT = _descriptor.Descriptor(
  name='Event',
  full_name='ray.rpc.Event',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='event_id', full_name='ray.rpc.Event.event_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='eventId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='source_type', full_name='ray.rpc.Event.source_type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='sourceType', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='source_hostname', full_name='ray.rpc.Event.source_hostname', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='sourceHostname', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='source_pid', full_name='ray.rpc.Event.source_pid', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='sourcePid', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='severity', full_name='ray.rpc.Event.severity', index=4,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='severity', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='label', full_name='ray.rpc.Event.label', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='label', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message', full_name='ray.rpc.Event.message', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='message', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='ray.rpc.Event.timestamp', index=7,
      number=8, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='timestamp', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='custom_fields', full_name='ray.rpc.Event.custom_fields', index=8,
      number=9, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='customFields', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_EVENT_CUSTOMFIELDSENTRY, ],
  enum_types=[
    _EVENT_SOURCETYPE,
    _EVENT_SEVERITY,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=42,
  serialized_end=596,
)


_REPORTEVENTSREPLY = _descriptor.Descriptor(
  name='ReportEventsReply',
  full_name='ray.rpc.ReportEventsReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='send_success', full_name='ray.rpc.ReportEventsReply.send_success', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='sendSuccess', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=598,
  serialized_end=652,
)


_REPORTEVENTSREQUEST = _descriptor.Descriptor(
  name='ReportEventsRequest',
  full_name='ray.rpc.ReportEventsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='event_strings', full_name='ray.rpc.ReportEventsRequest.event_strings', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='eventStrings', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=654,
  serialized_end=712,
)

_EVENT_CUSTOMFIELDSENTRY.containing_type = _EVENT
_EVENT.fields_by_name['source_type'].enum_type = _EVENT_SOURCETYPE
_EVENT.fields_by_name['severity'].enum_type = _EVENT_SEVERITY
_EVENT.fields_by_name['custom_fields'].message_type = _EVENT_CUSTOMFIELDSENTRY
_EVENT_SOURCETYPE.containing_type = _EVENT
_EVENT_SEVERITY.containing_type = _EVENT
DESCRIPTOR.message_types_by_name['Event'] = _EVENT
DESCRIPTOR.message_types_by_name['ReportEventsReply'] = _REPORTEVENTSREPLY
DESCRIPTOR.message_types_by_name['ReportEventsRequest'] = _REPORTEVENTSREQUEST
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Event = _reflection.GeneratedProtocolMessageType('Event', (_message.Message,), {

  'CustomFieldsEntry' : _reflection.GeneratedProtocolMessageType('CustomFieldsEntry', (_message.Message,), {
    'DESCRIPTOR' : _EVENT_CUSTOMFIELDSENTRY,
    '__module__' : 'src.ray.protobuf.event_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.Event.CustomFieldsEntry)
    })
  ,
  'DESCRIPTOR' : _EVENT,
  '__module__' : 'src.ray.protobuf.event_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.Event)
  })
_sym_db.RegisterMessage(Event)
_sym_db.RegisterMessage(Event.CustomFieldsEntry)

ReportEventsReply = _reflection.GeneratedProtocolMessageType('ReportEventsReply', (_message.Message,), {
  'DESCRIPTOR' : _REPORTEVENTSREPLY,
  '__module__' : 'src.ray.protobuf.event_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ReportEventsReply)
  })
_sym_db.RegisterMessage(ReportEventsReply)

ReportEventsRequest = _reflection.GeneratedProtocolMessageType('ReportEventsRequest', (_message.Message,), {
  'DESCRIPTOR' : _REPORTEVENTSREQUEST,
  '__module__' : 'src.ray.protobuf.event_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ReportEventsRequest)
  })
_sym_db.RegisterMessage(ReportEventsRequest)


DESCRIPTOR._options = None
_EVENT_CUSTOMFIELDSENTRY._options = None

_REPORTEVENTSERVICE = _descriptor.ServiceDescriptor(
  name='ReportEventService',
  full_name='ray.rpc.ReportEventService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=714,
  serialized_end=808,
  methods=[
  _descriptor.MethodDescriptor(
    name='ReportEvents',
    full_name='ray.rpc.ReportEventService.ReportEvents',
    index=0,
    containing_service=None,
    input_type=_REPORTEVENTSREQUEST,
    output_type=_REPORTEVENTSREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_REPORTEVENTSERVICE)

DESCRIPTOR.services_by_name['ReportEventService'] = _REPORTEVENTSERVICE

# @@protoc_insertion_point(module_scope)
