# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: src/ray/protobuf/pubsub.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import common_pb2 as src_dot_ray_dot_protobuf_dot_common__pb2
from . import gcs_pb2 as src_dot_ray_dot_protobuf_dot_gcs__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='src/ray/protobuf/pubsub.proto',
  package='ray.rpc',
  syntax='proto3',
  serialized_options=b'\370\001\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1dsrc/ray/protobuf/pubsub.proto\x12\x07ray.rpc\x1a\x1dsrc/ray/protobuf/common.proto\x1a\x1asrc/ray/protobuf/gcs.proto\"\x9f\x07\n\nPubMessage\x12\x37\n\x0c\x63hannel_type\x18\x01 \x01(\x0e\x32\x14.ray.rpc.ChannelTypeR\x0b\x63hannelType\x12\x15\n\x06key_id\x18\x02 \x01(\x0cR\x05keyId\x12k\n\x1eworker_object_eviction_message\x18\x03 \x01(\x0b\x32$.ray.rpc.WorkerObjectEvictionMessageH\x00R\x1bworkerObjectEvictionMessage\x12_\n\x1aworker_ref_removed_message\x18\x04 \x01(\x0b\x32 .ray.rpc.WorkerRefRemovedMessageH\x00R\x17workerRefRemovedMessage\x12q\n\x1fworker_object_locations_message\x18\x05 \x01(\x0b\x32(.ray.rpc.WorkerObjectLocationsPubMessageH\x00R\x1cworkerObjectLocationsMessage\x12>\n\ractor_message\x18\x07 \x01(\x0b\x32\x17.ray.rpc.ActorTableDataH\x00R\x0c\x61\x63torMessage\x12\x38\n\x0bjob_message\x18\x08 \x01(\x0b\x32\x15.ray.rpc.JobTableDataH\x00R\njobMessage\x12\x42\n\x11node_info_message\x18\t \x01(\x0b\x32\x14.ray.rpc.GcsNodeInfoH\x00R\x0fnodeInfoMessage\x12Q\n\x15node_resource_message\x18\n \x01(\x0b\x32\x1b.ray.rpc.NodeResourceChangeH\x00R\x13nodeResourceMessage\x12L\n\x14worker_delta_message\x18\x0b \x01(\x0b\x32\x18.ray.rpc.WorkerDeltaDataH\x00R\x12workerDeltaMessage\x12G\n\x12\x65rror_info_message\x18\x0c \x01(\x0b\x32\x17.ray.rpc.ErrorTableDataH\x00R\x10\x65rrorInfoMessage\x12\x42\n\x0f\x66\x61ilure_message\x18\x06 \x01(\x0b\x32\x17.ray.rpc.FailureMessageH\x00R\x0e\x66\x61ilureMessageB\x14\n\x12pub_message_one_of\":\n\x1bWorkerObjectEvictionMessage\x12\x1b\n\tobject_id\x18\x01 \x01(\x0cR\x08objectId\"]\n\x17WorkerRefRemovedMessage\x12\x42\n\rborrowed_refs\x18\x01 \x03(\x0b\x32\x1d.ray.rpc.ObjectReferenceCountR\x0c\x62orrowedRefs\"\xef\x01\n\x1fWorkerObjectLocationsPubMessage\x12\x19\n\x08node_ids\x18\x01 \x03(\x0cR\x07nodeIds\x12\x1f\n\x0bobject_size\x18\x02 \x01(\x04R\nobjectSize\x12\x1f\n\x0bspilled_url\x18\x03 \x01(\tR\nspilledUrl\x12&\n\x0fspilled_node_id\x18\x04 \x01(\x0cR\rspilledNodeId\x12&\n\x0fprimary_node_id\x18\x06 \x01(\x0cR\rprimaryNodeId\x12\x1f\n\x0bref_removed\x18\x07 \x01(\x08R\nrefRemoved\"\x10\n\x0e\x46\x61ilureMessage\"\x87\x02\n\x07\x43ommand\x12\x37\n\x0c\x63hannel_type\x18\x01 \x01(\x0e\x32\x14.ray.rpc.ChannelTypeR\x0b\x63hannelType\x12\x15\n\x06key_id\x18\x02 \x01(\x0cR\x05keyId\x12N\n\x13unsubscribe_message\x18\x03 \x01(\x0b\x32\x1b.ray.rpc.UnsubscribeMessageH\x00R\x12unsubscribeMessage\x12\x42\n\x11subscribe_message\x18\x04 \x01(\x0b\x32\x13.ray.rpc.SubMessageH\x00R\x10subscribeMessageB\x18\n\x16\x63ommand_message_one_of\"\x14\n\x12UnsubscribeMessage\"\xe9\x02\n\nSubMessage\x12n\n\x1eworker_object_eviction_message\x18\x01 \x01(\x0b\x32\'.ray.rpc.WorkerObjectEvictionSubMessageH\x00R\x1bworkerObjectEvictionMessage\x12\x62\n\x1aworker_ref_removed_message\x18\x02 \x01(\x0b\x32#.ray.rpc.WorkerRefRemovedSubMessageH\x00R\x17workerRefRemovedMessage\x12q\n\x1fworker_object_locations_message\x18\x03 \x01(\x0b\x32(.ray.rpc.WorkerObjectLocationsSubMessageH\x00R\x1cworkerObjectLocationsMessageB\x14\n\x12sub_message_one_of\"\xac\x01\n\x1eWorkerObjectEvictionSubMessage\x12,\n\x12intended_worker_id\x18\x01 \x01(\x0cR\x10intendedWorkerId\x12\x1b\n\tobject_id\x18\x02 \x01(\x0cR\x08objectId\x12?\n\x12subscriber_address\x18\x03 \x01(\x0b\x32\x10.ray.rpc.AddressR\x11subscriberAddress\"\xdc\x01\n\x1aWorkerRefRemovedSubMessage\x12,\n\x12intended_worker_id\x18\x01 \x01(\x0cR\x10intendedWorkerId\x12\x36\n\treference\x18\x02 \x01(\x0b\x32\x18.ray.rpc.ObjectReferenceR\treference\x12&\n\x0f\x63ontained_in_id\x18\x03 \x01(\x0cR\rcontainedInId\x12\x30\n\x14subscriber_worker_id\x18\x04 \x01(\x0cR\x12subscriberWorkerId\"l\n\x1fWorkerObjectLocationsSubMessage\x12,\n\x12intended_worker_id\x18\x01 \x01(\x0cR\x10intendedWorkerId\x12\x1b\n\tobject_id\x18\x02 \x01(\x0cR\x08objectId\"?\n\x18PubsubLongPollingRequest\x12#\n\rsubscriber_id\x18\x01 \x01(\x0cR\x0csubscriberId\"P\n\x16PubsubLongPollingReply\x12\x36\n\x0cpub_messages\x18\x01 \x03(\x0b\x32\x13.ray.rpc.PubMessageR\x0bpubMessages\"n\n\x19PubsubCommandBatchRequest\x12#\n\rsubscriber_id\x18\x01 \x01(\x0cR\x0csubscriberId\x12,\n\x08\x63ommands\x18\x02 \x03(\x0b\x32\x10.ray.rpc.CommandR\x08\x63ommands\"\x19\n\x17PubsubCommandBatchReply*\x8e\x02\n\x0b\x43hannelType\x12\x1a\n\x16WORKER_OBJECT_EVICTION\x10\x00\x12\x1e\n\x1aWORKER_REF_REMOVED_CHANNEL\x10\x01\x12#\n\x1fWORKER_OBJECT_LOCATIONS_CHANNEL\x10\x02\x12\x15\n\x11GCS_ACTOR_CHANNEL\x10\x03\x12\x13\n\x0fGCS_JOB_CHANNEL\x10\x04\x12\x19\n\x15GCS_NODE_INFO_CHANNEL\x10\x05\x12\x1d\n\x19GCS_NODE_RESOURCE_CHANNEL\x10\x06\x12\x1c\n\x18GCS_WORKER_DELTA_CHANNEL\x10\x07\x12\x1a\n\x16RAY_ERROR_INFO_CHANNEL\x10\x08\x32\xc7\x01\n\x10PublisherService\x12W\n\x11PubsubLongPolling\x12!.ray.rpc.PubsubLongPollingRequest\x1a\x1f.ray.rpc.PubsubLongPollingReply\x12Z\n\x12PubsubCommandBatch\x12\".ray.rpc.PubsubCommandBatchRequest\x1a .ray.rpc.PubsubCommandBatchReplyB\x03\xf8\x01\x01\x62\x06proto3'
  ,
  dependencies=[src_dot_ray_dot_protobuf_dot_common__pb2.DESCRIPTOR,src_dot_ray_dot_protobuf_dot_gcs__pb2.DESCRIPTOR,])

_CHANNELTYPE = _descriptor.EnumDescriptor(
  name='ChannelType',
  full_name='ray.rpc.ChannelType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='WORKER_OBJECT_EVICTION', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='WORKER_REF_REMOVED_CHANNEL', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='WORKER_OBJECT_LOCATIONS_CHANNEL', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='GCS_ACTOR_CHANNEL', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='GCS_JOB_CHANNEL', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='GCS_NODE_INFO_CHANNEL', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='GCS_NODE_RESOURCE_CHANNEL', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='GCS_WORKER_DELTA_CHANNEL', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='RAY_ERROR_INFO_CHANNEL', index=8, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2893,
  serialized_end=3163,
)
_sym_db.RegisterEnumDescriptor(_CHANNELTYPE)

ChannelType = enum_type_wrapper.EnumTypeWrapper(_CHANNELTYPE)
WORKER_OBJECT_EVICTION = 0
WORKER_REF_REMOVED_CHANNEL = 1
WORKER_OBJECT_LOCATIONS_CHANNEL = 2
GCS_ACTOR_CHANNEL = 3
GCS_JOB_CHANNEL = 4
GCS_NODE_INFO_CHANNEL = 5
GCS_NODE_RESOURCE_CHANNEL = 6
GCS_WORKER_DELTA_CHANNEL = 7
RAY_ERROR_INFO_CHANNEL = 8



_PUBMESSAGE = _descriptor.Descriptor(
  name='PubMessage',
  full_name='ray.rpc.PubMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='channel_type', full_name='ray.rpc.PubMessage.channel_type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='channelType', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='key_id', full_name='ray.rpc.PubMessage.key_id', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='keyId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='worker_object_eviction_message', full_name='ray.rpc.PubMessage.worker_object_eviction_message', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='workerObjectEvictionMessage', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='worker_ref_removed_message', full_name='ray.rpc.PubMessage.worker_ref_removed_message', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='workerRefRemovedMessage', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='worker_object_locations_message', full_name='ray.rpc.PubMessage.worker_object_locations_message', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='workerObjectLocationsMessage', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='actor_message', full_name='ray.rpc.PubMessage.actor_message', index=5,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='actorMessage', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='job_message', full_name='ray.rpc.PubMessage.job_message', index=6,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='jobMessage', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='node_info_message', full_name='ray.rpc.PubMessage.node_info_message', index=7,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='nodeInfoMessage', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='node_resource_message', full_name='ray.rpc.PubMessage.node_resource_message', index=8,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='nodeResourceMessage', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='worker_delta_message', full_name='ray.rpc.PubMessage.worker_delta_message', index=9,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='workerDeltaMessage', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='error_info_message', full_name='ray.rpc.PubMessage.error_info_message', index=10,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='errorInfoMessage', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='failure_message', full_name='ray.rpc.PubMessage.failure_message', index=11,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='failureMessage', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='pub_message_one_of', full_name='ray.rpc.PubMessage.pub_message_one_of',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=102,
  serialized_end=1029,
)


_WORKEROBJECTEVICTIONMESSAGE = _descriptor.Descriptor(
  name='WorkerObjectEvictionMessage',
  full_name='ray.rpc.WorkerObjectEvictionMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='object_id', full_name='ray.rpc.WorkerObjectEvictionMessage.object_id', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='objectId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1031,
  serialized_end=1089,
)


_WORKERREFREMOVEDMESSAGE = _descriptor.Descriptor(
  name='WorkerRefRemovedMessage',
  full_name='ray.rpc.WorkerRefRemovedMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='borrowed_refs', full_name='ray.rpc.WorkerRefRemovedMessage.borrowed_refs', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='borrowedRefs', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1091,
  serialized_end=1184,
)


_WORKEROBJECTLOCATIONSPUBMESSAGE = _descriptor.Descriptor(
  name='WorkerObjectLocationsPubMessage',
  full_name='ray.rpc.WorkerObjectLocationsPubMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='node_ids', full_name='ray.rpc.WorkerObjectLocationsPubMessage.node_ids', index=0,
      number=1, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='nodeIds', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='object_size', full_name='ray.rpc.WorkerObjectLocationsPubMessage.object_size', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='objectSize', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='spilled_url', full_name='ray.rpc.WorkerObjectLocationsPubMessage.spilled_url', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='spilledUrl', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='spilled_node_id', full_name='ray.rpc.WorkerObjectLocationsPubMessage.spilled_node_id', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='spilledNodeId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='primary_node_id', full_name='ray.rpc.WorkerObjectLocationsPubMessage.primary_node_id', index=4,
      number=6, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='primaryNodeId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ref_removed', full_name='ray.rpc.WorkerObjectLocationsPubMessage.ref_removed', index=5,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='refRemoved', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1187,
  serialized_end=1426,
)


_FAILUREMESSAGE = _descriptor.Descriptor(
  name='FailureMessage',
  full_name='ray.rpc.FailureMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1428,
  serialized_end=1444,
)


_COMMAND = _descriptor.Descriptor(
  name='Command',
  full_name='ray.rpc.Command',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='channel_type', full_name='ray.rpc.Command.channel_type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='channelType', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='key_id', full_name='ray.rpc.Command.key_id', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='keyId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='unsubscribe_message', full_name='ray.rpc.Command.unsubscribe_message', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='unsubscribeMessage', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='subscribe_message', full_name='ray.rpc.Command.subscribe_message', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='subscribeMessage', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='command_message_one_of', full_name='ray.rpc.Command.command_message_one_of',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=1447,
  serialized_end=1710,
)


_UNSUBSCRIBEMESSAGE = _descriptor.Descriptor(
  name='UnsubscribeMessage',
  full_name='ray.rpc.UnsubscribeMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1712,
  serialized_end=1732,
)


_SUBMESSAGE = _descriptor.Descriptor(
  name='SubMessage',
  full_name='ray.rpc.SubMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='worker_object_eviction_message', full_name='ray.rpc.SubMessage.worker_object_eviction_message', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='workerObjectEvictionMessage', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='worker_ref_removed_message', full_name='ray.rpc.SubMessage.worker_ref_removed_message', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='workerRefRemovedMessage', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='worker_object_locations_message', full_name='ray.rpc.SubMessage.worker_object_locations_message', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='workerObjectLocationsMessage', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='sub_message_one_of', full_name='ray.rpc.SubMessage.sub_message_one_of',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=1735,
  serialized_end=2096,
)


_WORKEROBJECTEVICTIONSUBMESSAGE = _descriptor.Descriptor(
  name='WorkerObjectEvictionSubMessage',
  full_name='ray.rpc.WorkerObjectEvictionSubMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='intended_worker_id', full_name='ray.rpc.WorkerObjectEvictionSubMessage.intended_worker_id', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='intendedWorkerId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='object_id', full_name='ray.rpc.WorkerObjectEvictionSubMessage.object_id', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='objectId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='subscriber_address', full_name='ray.rpc.WorkerObjectEvictionSubMessage.subscriber_address', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='subscriberAddress', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2099,
  serialized_end=2271,
)


_WORKERREFREMOVEDSUBMESSAGE = _descriptor.Descriptor(
  name='WorkerRefRemovedSubMessage',
  full_name='ray.rpc.WorkerRefRemovedSubMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='intended_worker_id', full_name='ray.rpc.WorkerRefRemovedSubMessage.intended_worker_id', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='intendedWorkerId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='reference', full_name='ray.rpc.WorkerRefRemovedSubMessage.reference', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='reference', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='contained_in_id', full_name='ray.rpc.WorkerRefRemovedSubMessage.contained_in_id', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='containedInId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='subscriber_worker_id', full_name='ray.rpc.WorkerRefRemovedSubMessage.subscriber_worker_id', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='subscriberWorkerId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2274,
  serialized_end=2494,
)


_WORKEROBJECTLOCATIONSSUBMESSAGE = _descriptor.Descriptor(
  name='WorkerObjectLocationsSubMessage',
  full_name='ray.rpc.WorkerObjectLocationsSubMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='intended_worker_id', full_name='ray.rpc.WorkerObjectLocationsSubMessage.intended_worker_id', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='intendedWorkerId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='object_id', full_name='ray.rpc.WorkerObjectLocationsSubMessage.object_id', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='objectId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2496,
  serialized_end=2604,
)


_PUBSUBLONGPOLLINGREQUEST = _descriptor.Descriptor(
  name='PubsubLongPollingRequest',
  full_name='ray.rpc.PubsubLongPollingRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='subscriber_id', full_name='ray.rpc.PubsubLongPollingRequest.subscriber_id', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='subscriberId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2606,
  serialized_end=2669,
)


_PUBSUBLONGPOLLINGREPLY = _descriptor.Descriptor(
  name='PubsubLongPollingReply',
  full_name='ray.rpc.PubsubLongPollingReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='pub_messages', full_name='ray.rpc.PubsubLongPollingReply.pub_messages', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='pubMessages', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2671,
  serialized_end=2751,
)


_PUBSUBCOMMANDBATCHREQUEST = _descriptor.Descriptor(
  name='PubsubCommandBatchRequest',
  full_name='ray.rpc.PubsubCommandBatchRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='subscriber_id', full_name='ray.rpc.PubsubCommandBatchRequest.subscriber_id', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='subscriberId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='commands', full_name='ray.rpc.PubsubCommandBatchRequest.commands', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='commands', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2753,
  serialized_end=2863,
)


_PUBSUBCOMMANDBATCHREPLY = _descriptor.Descriptor(
  name='PubsubCommandBatchReply',
  full_name='ray.rpc.PubsubCommandBatchReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2865,
  serialized_end=2890,
)

_PUBMESSAGE.fields_by_name['channel_type'].enum_type = _CHANNELTYPE
_PUBMESSAGE.fields_by_name['worker_object_eviction_message'].message_type = _WORKEROBJECTEVICTIONMESSAGE
_PUBMESSAGE.fields_by_name['worker_ref_removed_message'].message_type = _WORKERREFREMOVEDMESSAGE
_PUBMESSAGE.fields_by_name['worker_object_locations_message'].message_type = _WORKEROBJECTLOCATIONSPUBMESSAGE
_PUBMESSAGE.fields_by_name['actor_message'].message_type = src_dot_ray_dot_protobuf_dot_gcs__pb2._ACTORTABLEDATA
_PUBMESSAGE.fields_by_name['job_message'].message_type = src_dot_ray_dot_protobuf_dot_gcs__pb2._JOBTABLEDATA
_PUBMESSAGE.fields_by_name['node_info_message'].message_type = src_dot_ray_dot_protobuf_dot_gcs__pb2._GCSNODEINFO
_PUBMESSAGE.fields_by_name['node_resource_message'].message_type = src_dot_ray_dot_protobuf_dot_gcs__pb2._NODERESOURCECHANGE
_PUBMESSAGE.fields_by_name['worker_delta_message'].message_type = src_dot_ray_dot_protobuf_dot_gcs__pb2._WORKERDELTADATA
_PUBMESSAGE.fields_by_name['error_info_message'].message_type = src_dot_ray_dot_protobuf_dot_gcs__pb2._ERRORTABLEDATA
_PUBMESSAGE.fields_by_name['failure_message'].message_type = _FAILUREMESSAGE
_PUBMESSAGE.oneofs_by_name['pub_message_one_of'].fields.append(
  _PUBMESSAGE.fields_by_name['worker_object_eviction_message'])
_PUBMESSAGE.fields_by_name['worker_object_eviction_message'].containing_oneof = _PUBMESSAGE.oneofs_by_name['pub_message_one_of']
_PUBMESSAGE.oneofs_by_name['pub_message_one_of'].fields.append(
  _PUBMESSAGE.fields_by_name['worker_ref_removed_message'])
_PUBMESSAGE.fields_by_name['worker_ref_removed_message'].containing_oneof = _PUBMESSAGE.oneofs_by_name['pub_message_one_of']
_PUBMESSAGE.oneofs_by_name['pub_message_one_of'].fields.append(
  _PUBMESSAGE.fields_by_name['worker_object_locations_message'])
_PUBMESSAGE.fields_by_name['worker_object_locations_message'].containing_oneof = _PUBMESSAGE.oneofs_by_name['pub_message_one_of']
_PUBMESSAGE.oneofs_by_name['pub_message_one_of'].fields.append(
  _PUBMESSAGE.fields_by_name['actor_message'])
_PUBMESSAGE.fields_by_name['actor_message'].containing_oneof = _PUBMESSAGE.oneofs_by_name['pub_message_one_of']
_PUBMESSAGE.oneofs_by_name['pub_message_one_of'].fields.append(
  _PUBMESSAGE.fields_by_name['job_message'])
_PUBMESSAGE.fields_by_name['job_message'].containing_oneof = _PUBMESSAGE.oneofs_by_name['pub_message_one_of']
_PUBMESSAGE.oneofs_by_name['pub_message_one_of'].fields.append(
  _PUBMESSAGE.fields_by_name['node_info_message'])
_PUBMESSAGE.fields_by_name['node_info_message'].containing_oneof = _PUBMESSAGE.oneofs_by_name['pub_message_one_of']
_PUBMESSAGE.oneofs_by_name['pub_message_one_of'].fields.append(
  _PUBMESSAGE.fields_by_name['node_resource_message'])
_PUBMESSAGE.fields_by_name['node_resource_message'].containing_oneof = _PUBMESSAGE.oneofs_by_name['pub_message_one_of']
_PUBMESSAGE.oneofs_by_name['pub_message_one_of'].fields.append(
  _PUBMESSAGE.fields_by_name['worker_delta_message'])
_PUBMESSAGE.fields_by_name['worker_delta_message'].containing_oneof = _PUBMESSAGE.oneofs_by_name['pub_message_one_of']
_PUBMESSAGE.oneofs_by_name['pub_message_one_of'].fields.append(
  _PUBMESSAGE.fields_by_name['error_info_message'])
_PUBMESSAGE.fields_by_name['error_info_message'].containing_oneof = _PUBMESSAGE.oneofs_by_name['pub_message_one_of']
_PUBMESSAGE.oneofs_by_name['pub_message_one_of'].fields.append(
  _PUBMESSAGE.fields_by_name['failure_message'])
_PUBMESSAGE.fields_by_name['failure_message'].containing_oneof = _PUBMESSAGE.oneofs_by_name['pub_message_one_of']
_WORKERREFREMOVEDMESSAGE.fields_by_name['borrowed_refs'].message_type = src_dot_ray_dot_protobuf_dot_common__pb2._OBJECTREFERENCECOUNT
_COMMAND.fields_by_name['channel_type'].enum_type = _CHANNELTYPE
_COMMAND.fields_by_name['unsubscribe_message'].message_type = _UNSUBSCRIBEMESSAGE
_COMMAND.fields_by_name['subscribe_message'].message_type = _SUBMESSAGE
_COMMAND.oneofs_by_name['command_message_one_of'].fields.append(
  _COMMAND.fields_by_name['unsubscribe_message'])
_COMMAND.fields_by_name['unsubscribe_message'].containing_oneof = _COMMAND.oneofs_by_name['command_message_one_of']
_COMMAND.oneofs_by_name['command_message_one_of'].fields.append(
  _COMMAND.fields_by_name['subscribe_message'])
_COMMAND.fields_by_name['subscribe_message'].containing_oneof = _COMMAND.oneofs_by_name['command_message_one_of']
_SUBMESSAGE.fields_by_name['worker_object_eviction_message'].message_type = _WORKEROBJECTEVICTIONSUBMESSAGE
_SUBMESSAGE.fields_by_name['worker_ref_removed_message'].message_type = _WORKERREFREMOVEDSUBMESSAGE
_SUBMESSAGE.fields_by_name['worker_object_locations_message'].message_type = _WORKEROBJECTLOCATIONSSUBMESSAGE
_SUBMESSAGE.oneofs_by_name['sub_message_one_of'].fields.append(
  _SUBMESSAGE.fields_by_name['worker_object_eviction_message'])
_SUBMESSAGE.fields_by_name['worker_object_eviction_message'].containing_oneof = _SUBMESSAGE.oneofs_by_name['sub_message_one_of']
_SUBMESSAGE.oneofs_by_name['sub_message_one_of'].fields.append(
  _SUBMESSAGE.fields_by_name['worker_ref_removed_message'])
_SUBMESSAGE.fields_by_name['worker_ref_removed_message'].containing_oneof = _SUBMESSAGE.oneofs_by_name['sub_message_one_of']
_SUBMESSAGE.oneofs_by_name['sub_message_one_of'].fields.append(
  _SUBMESSAGE.fields_by_name['worker_object_locations_message'])
_SUBMESSAGE.fields_by_name['worker_object_locations_message'].containing_oneof = _SUBMESSAGE.oneofs_by_name['sub_message_one_of']
_WORKEROBJECTEVICTIONSUBMESSAGE.fields_by_name['subscriber_address'].message_type = src_dot_ray_dot_protobuf_dot_common__pb2._ADDRESS
_WORKERREFREMOVEDSUBMESSAGE.fields_by_name['reference'].message_type = src_dot_ray_dot_protobuf_dot_common__pb2._OBJECTREFERENCE
_PUBSUBLONGPOLLINGREPLY.fields_by_name['pub_messages'].message_type = _PUBMESSAGE
_PUBSUBCOMMANDBATCHREQUEST.fields_by_name['commands'].message_type = _COMMAND
DESCRIPTOR.message_types_by_name['PubMessage'] = _PUBMESSAGE
DESCRIPTOR.message_types_by_name['WorkerObjectEvictionMessage'] = _WORKEROBJECTEVICTIONMESSAGE
DESCRIPTOR.message_types_by_name['WorkerRefRemovedMessage'] = _WORKERREFREMOVEDMESSAGE
DESCRIPTOR.message_types_by_name['WorkerObjectLocationsPubMessage'] = _WORKEROBJECTLOCATIONSPUBMESSAGE
DESCRIPTOR.message_types_by_name['FailureMessage'] = _FAILUREMESSAGE
DESCRIPTOR.message_types_by_name['Command'] = _COMMAND
DESCRIPTOR.message_types_by_name['UnsubscribeMessage'] = _UNSUBSCRIBEMESSAGE
DESCRIPTOR.message_types_by_name['SubMessage'] = _SUBMESSAGE
DESCRIPTOR.message_types_by_name['WorkerObjectEvictionSubMessage'] = _WORKEROBJECTEVICTIONSUBMESSAGE
DESCRIPTOR.message_types_by_name['WorkerRefRemovedSubMessage'] = _WORKERREFREMOVEDSUBMESSAGE
DESCRIPTOR.message_types_by_name['WorkerObjectLocationsSubMessage'] = _WORKEROBJECTLOCATIONSSUBMESSAGE
DESCRIPTOR.message_types_by_name['PubsubLongPollingRequest'] = _PUBSUBLONGPOLLINGREQUEST
DESCRIPTOR.message_types_by_name['PubsubLongPollingReply'] = _PUBSUBLONGPOLLINGREPLY
DESCRIPTOR.message_types_by_name['PubsubCommandBatchRequest'] = _PUBSUBCOMMANDBATCHREQUEST
DESCRIPTOR.message_types_by_name['PubsubCommandBatchReply'] = _PUBSUBCOMMANDBATCHREPLY
DESCRIPTOR.enum_types_by_name['ChannelType'] = _CHANNELTYPE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

PubMessage = _reflection.GeneratedProtocolMessageType('PubMessage', (_message.Message,), {
  'DESCRIPTOR' : _PUBMESSAGE,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PubMessage)
  })
_sym_db.RegisterMessage(PubMessage)

WorkerObjectEvictionMessage = _reflection.GeneratedProtocolMessageType('WorkerObjectEvictionMessage', (_message.Message,), {
  'DESCRIPTOR' : _WORKEROBJECTEVICTIONMESSAGE,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.WorkerObjectEvictionMessage)
  })
_sym_db.RegisterMessage(WorkerObjectEvictionMessage)

WorkerRefRemovedMessage = _reflection.GeneratedProtocolMessageType('WorkerRefRemovedMessage', (_message.Message,), {
  'DESCRIPTOR' : _WORKERREFREMOVEDMESSAGE,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.WorkerRefRemovedMessage)
  })
_sym_db.RegisterMessage(WorkerRefRemovedMessage)

WorkerObjectLocationsPubMessage = _reflection.GeneratedProtocolMessageType('WorkerObjectLocationsPubMessage', (_message.Message,), {
  'DESCRIPTOR' : _WORKEROBJECTLOCATIONSPUBMESSAGE,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.WorkerObjectLocationsPubMessage)
  })
_sym_db.RegisterMessage(WorkerObjectLocationsPubMessage)

FailureMessage = _reflection.GeneratedProtocolMessageType('FailureMessage', (_message.Message,), {
  'DESCRIPTOR' : _FAILUREMESSAGE,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.FailureMessage)
  })
_sym_db.RegisterMessage(FailureMessage)

Command = _reflection.GeneratedProtocolMessageType('Command', (_message.Message,), {
  'DESCRIPTOR' : _COMMAND,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.Command)
  })
_sym_db.RegisterMessage(Command)

UnsubscribeMessage = _reflection.GeneratedProtocolMessageType('UnsubscribeMessage', (_message.Message,), {
  'DESCRIPTOR' : _UNSUBSCRIBEMESSAGE,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.UnsubscribeMessage)
  })
_sym_db.RegisterMessage(UnsubscribeMessage)

SubMessage = _reflection.GeneratedProtocolMessageType('SubMessage', (_message.Message,), {
  'DESCRIPTOR' : _SUBMESSAGE,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.SubMessage)
  })
_sym_db.RegisterMessage(SubMessage)

WorkerObjectEvictionSubMessage = _reflection.GeneratedProtocolMessageType('WorkerObjectEvictionSubMessage', (_message.Message,), {
  'DESCRIPTOR' : _WORKEROBJECTEVICTIONSUBMESSAGE,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.WorkerObjectEvictionSubMessage)
  })
_sym_db.RegisterMessage(WorkerObjectEvictionSubMessage)

WorkerRefRemovedSubMessage = _reflection.GeneratedProtocolMessageType('WorkerRefRemovedSubMessage', (_message.Message,), {
  'DESCRIPTOR' : _WORKERREFREMOVEDSUBMESSAGE,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.WorkerRefRemovedSubMessage)
  })
_sym_db.RegisterMessage(WorkerRefRemovedSubMessage)

WorkerObjectLocationsSubMessage = _reflection.GeneratedProtocolMessageType('WorkerObjectLocationsSubMessage', (_message.Message,), {
  'DESCRIPTOR' : _WORKEROBJECTLOCATIONSSUBMESSAGE,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.WorkerObjectLocationsSubMessage)
  })
_sym_db.RegisterMessage(WorkerObjectLocationsSubMessage)

PubsubLongPollingRequest = _reflection.GeneratedProtocolMessageType('PubsubLongPollingRequest', (_message.Message,), {
  'DESCRIPTOR' : _PUBSUBLONGPOLLINGREQUEST,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PubsubLongPollingRequest)
  })
_sym_db.RegisterMessage(PubsubLongPollingRequest)

PubsubLongPollingReply = _reflection.GeneratedProtocolMessageType('PubsubLongPollingReply', (_message.Message,), {
  'DESCRIPTOR' : _PUBSUBLONGPOLLINGREPLY,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PubsubLongPollingReply)
  })
_sym_db.RegisterMessage(PubsubLongPollingReply)

PubsubCommandBatchRequest = _reflection.GeneratedProtocolMessageType('PubsubCommandBatchRequest', (_message.Message,), {
  'DESCRIPTOR' : _PUBSUBCOMMANDBATCHREQUEST,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PubsubCommandBatchRequest)
  })
_sym_db.RegisterMessage(PubsubCommandBatchRequest)

PubsubCommandBatchReply = _reflection.GeneratedProtocolMessageType('PubsubCommandBatchReply', (_message.Message,), {
  'DESCRIPTOR' : _PUBSUBCOMMANDBATCHREPLY,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PubsubCommandBatchReply)
  })
_sym_db.RegisterMessage(PubsubCommandBatchReply)


DESCRIPTOR._options = None

_PUBLISHERSERVICE = _descriptor.ServiceDescriptor(
  name='PublisherService',
  full_name='ray.rpc.PublisherService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=3166,
  serialized_end=3365,
  methods=[
  _descriptor.MethodDescriptor(
    name='PubsubLongPolling',
    full_name='ray.rpc.PublisherService.PubsubLongPolling',
    index=0,
    containing_service=None,
    input_type=_PUBSUBLONGPOLLINGREQUEST,
    output_type=_PUBSUBLONGPOLLINGREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='PubsubCommandBatch',
    full_name='ray.rpc.PublisherService.PubsubCommandBatch',
    index=1,
    containing_service=None,
    input_type=_PUBSUBCOMMANDBATCHREQUEST,
    output_type=_PUBSUBCOMMANDBATCHREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_PUBLISHERSERVICE)

DESCRIPTOR.services_by_name['PublisherService'] = _PUBLISHERSERVICE

# @@protoc_insertion_point(module_scope)
