# List of files to exclude from the Ray directory when using runtime_env for
# Ray development. These are not necessary in the Ray workers.
RAY_WORKER_DEV_EXCLUDES = [
    "raylet", "gcs_server", "cpp/", "tests/", "core/src"
]
